/* mare/timer.c */
/* Functions for calculating resource/cpu time used by TinyMARE.
   All `quad' values are in milliseconds unless otherwise stated. */

#include "externs.h"
#include <sys/resource.h>

dbref object_atime=NOTHING;

long dump_interval;  // Seconds (ticks) since last dump.
int cpu_percent;     // CPU Usage in the last 60 seconds (95000 = 95%).

double cmdfp[2];     // Running #Cmds/sec averages.
int cmdav[2];        // #Cmds at last checkpoint.


/* Get current system time (resolution in milliseconds) */
quad get_time()
{
  struct timeval tv;

  gettimeofday(&tv, NULL);
  return (quad)tv.tv_sec*1000 + tv.tv_usec/1000;
}

/* Returns the amount of CPU time used */
quad get_cputime()
{
  struct rusage usage;

  getrusage(RUSAGE_SELF, &usage);
  return (quad)(usage.ru_utime.tv_sec+usage.ru_stime.tv_sec)*1000 +
          (usage.ru_utime.tv_usec+usage.ru_stime.tv_usec)/1000;
}

/* Called by the timer every 60 seconds to update the global 'cpu_percent' */
void update_cpu_percent()
{
  static quad last_cpu, last_time;
  quad cur_cpu, cur_time;

  cur_time=get_time();
  cur_cpu=get_cputime();

  /* Record the last minute's CPU% usage */
  if(last_time && last_time < cur_time)
    cpu_percent=(cur_cpu-last_cpu)*100000/(cur_time-last_time);
  else
    cpu_percent=0;

  last_time=cur_time;
  last_cpu=cur_cpu;
}

static void next_atime()
{
  static int regen_loop=0, wander_loop=0;
  char *s;
  int a, i;

  if(ATIME_INTERVAL < 1)
    return;

  for(i=db_top/ATIME_INTERVAL;i >= 0;i--) {
    /* when @atime reaches the last object in the database,
       loop around to the beginning. */
    if(++object_atime >= db_top) {
      object_atime = NOTHING;
      if(++regen_loop >= RESURRECT_LOOP)
        regen_loop=0;
      if(++wander_loop >= WANDER_LOOP)
        wander_loop=0;
      return;
    }

    /* skip over destroyed objects or disconnected players */
    if((db[object_atime].flags & (GOING|HAVEN)) || (Typeof(object_atime) ==
       TYPE_PLAYER && !(db[object_atime].flags & PLAYER_CONNECT)))
      continue;

    /* set %# for messages */
    speaker=object_atime;

    /* execute @atime */
    if(*(s=atr_get(object_atime, A_ATIME))) {
      immediate_que(object_atime, s, object_atime);
      for(a=0;a<10;a++)
        *wptr[a]='\0';
    }

    /* Objects set WANDER occasionally move around */
    if(IS(object_atime, TYPE_THING, THING_WANDER) &&
       WANDER_LOOP > 0 && (object_atime % WANDER_LOOP) == wander_loop) {
      a=random_exit(object_atime, NOTHING);
      if(!Invalid(a))
        do_move(object_atime, db[a].name);
    }
  }
}


/* Timer Functions */

// do_decisecond(): Triggers every 0.1 seconds.
//  ticks := Number of 0.1-second intervals elapsed since last call.
//
static void do_decisecond(int ticks)
{
}

// do_second(): Triggers once per second.
//  ticks := Number of seconds elapsed since last call.
//
static void do_second(int ticks)
{
  static int cmdav_loop=0;
  DESC *d;

  /* update players' online status */
  for(d=descriptor_list;d;d=d->next) {
    if((d->flags & C_CONNECTED) && !IS(db[d->player].location, TYPE_ROOM,
       ROOM_XZONE))
      db[d->player].data->age+=ticks;
  }

  /* other stuff to do every second */

#ifndef __CYGWIN__
  dns_events();			// Keep track of expired dns cache records
#endif

  update_io();			// Updates login prompt and input quotas
  purge_going_objs();		// Purge any objects scheduled for destruction
  wait_second();		// Process queue of @waited commands

  next_atime();			// Run the next set of @atimes

  /* Update running command averages */
  if((cmdav_loop += ticks) >= 5) {
    cmdfp[0]+=(((double)(nstat[NS_NPCMDS]-cmdav[0])/cmdav_loop)-cmdfp[0])/3;
    cmdfp[1]+=(((double)(nstat[NS_NQCMDS]-cmdav[1])/cmdav_loop)-cmdfp[1])/3;
    cmdav[0]=nstat[NS_NPCMDS];
    cmdav[1]=nstat[NS_NQCMDS];
    cmdav_loop=0;
  }

  /* Automatic Database Saving */
  if(!halt_dump && !db_readonly) {
    dump_interval += ticks;
    if(dump_interval > DUMP_INTERVAL) {
      dump_interval=0;
      fork_and_dump();
    }
  }
}

// do_minute(): Triggers each minute on the minute.
//
static void do_minute()
{
  DESC *d;
  static int last_weekday=-1, report_loop;
  int weekday;

  /* loop for all connected sockets */
  for(d=descriptor_list;d;d=d->next) {
    /* ping concentrators every minute */
    if(d->flags & C_PARENT)
      output2(d, "*\n");

    /* ping remotelink services */
    else if((d->flags & C_SERVICE) && d->state == 33) {
      DESC *k;
      int a;

      for(a=0,k=descriptor_list;k;k=k->next)
        if((k->flags & C_CONNECTED) && could_doit(0, k->player, A_HLOCK))
          a++;
      output2(d, tprintf("$20 %d\n", a));
    }

    /* process session inactivity timeout */
    else if(!(d->flags & (C_OUTBOUND|C_SOCKET|C_DNS|C_PARENT|C_SERVICE)))
      check_timeout(d);
  }

  /* Report disconnected rooms every 10 minutes */
  if(!(++report_loop % 10))
    report_disconnected_rooms(NOTHING);

  /* other stuff to do every minute */

  update_cpu_percent();		// Calculate CPU% Usage used in past 60sec

  /* Set weekday to 0..6, 0 being Sunday */
  weekday=(((now+(TIMEZONE*3600))/86400)-3) % 7;

  /* The following executes daily at midnight */
  if(last_weekday == -1)
    last_weekday=weekday;
  else if(last_weekday != weekday) {
    last_weekday=weekday;

    // Remove old players from the game every Sunday at midnight.
    if(INACTIVENUKE > 0 && weekday == 0) {
      log_main("Invoking inactivity deletion...");
      inactive_nuke();
    }
  }
}


/* Main Timer Loop */

void dispatch()
{
  static quad deci, second, minute;  /* Timers */
  quad current=get_time(), timer;
  int ticks;

  /* Deciseconds */
  timer=current/100;
  if(timer == deci)
    return;
  ticks=timer-deci;
  deci=timer;
  if(ticks < 0 || ticks > 20)
    ticks=1;
  do_decisecond(ticks);

  /* Seconds */
  timer/=10;
  if(timer == second)
    return;
  ticks=timer-second;
  second=timer;
  if(ticks < 0 || ticks > 20)
    return;
  do_second(ticks);

  /* Minutes */
  timer/=60;
  if(timer == minute)
    return;
  if(minute)
    do_minute();
  minute=timer;
}
